from typing import Any, Dict, Tuple, List, DefaultDict
from collections import defaultdict


from retrying import retry  # type: ignore

from capella_console_client.logconf import logger
from capella_console_client.session import CapellaConsoleSession
from capella_console_client.config import (
    ALL_SUPPORTED_FIELDS,
    ALL_SUPPORTED_SORTBY,
    SUPPORTED_SEARCH_FIELDS,
    SUPPORTED_QUERY_FIELDS,
    STAC_PREFIXED_BY_QUERY_FIELDS,
    OPERATOR_SUFFIXES,
    DEFAULT_PAGE_SIZE,
    DEFAULT_MAX_FEATURE_COUNT,
)
from capella_console_client.hooks import retry_if_http_status_error


def _build_search_payload(**kwargs) -> Dict[str, Any]:
    payload = {}
    query_payload: DefaultDict[str, Dict[str, Any]] = defaultdict(dict)

    for field, value in kwargs.items():
        field, op = _split_op(field)
        if field not in ALL_SUPPORTED_FIELDS:
            logger.warning(f"filter {field} not supported ... omitting")
            continue

        if op not in OPERATOR_SUFFIXES:
            logger.warning(f"operator {op} not supported ... omitting")
            continue

        if field in SUPPORTED_SEARCH_FIELDS:
            payload[field] = value
        elif field in SUPPORTED_QUERY_FIELDS:
            if type(value) == list:
                op = "in"

            target_field = STAC_PREFIXED_BY_QUERY_FIELDS.get(field, field)
            query_payload[target_field][op] = value

    if query_payload:
        payload["query"] = dict(query_payload)

    if "sortby" in kwargs:
        payload["sortby"] = _get_sort_payload(kwargs)

    return payload


def _get_sort_payload(kwargs):
    directions = {"-": "desc", "+": "asc"}
    sorts = []
    orig = kwargs["sortby"]

    if not isinstance(orig, list):
        orig = [orig]

    for sort_arg in orig:
        field = sort_arg[1:]
        direction = sort_arg[0]
        if direction not in directions:
            direction = "+"
            field = sort_arg

        if field not in ALL_SUPPORTED_SORTBY:
            logger.warning(f"sorting by {field} not supported ... omitting")
            continue

        if field in SUPPORTED_QUERY_FIELDS or field == "datetime":
            field = f"properties.{field}"

        sorts.append({"field": field, "direction": directions[direction]})
    return sorts


def _split_op(field: str) -> Tuple[str, str]:
    parts = field.split("__")
    if len(parts) == 2:
        op = parts[1]
    else:
        op = "eq"
    return (parts[0], op)


def _paginated_search(
    session: CapellaConsoleSession, payload: Dict[str, Any]
) -> List[Dict[str, Any]]:
    max_cnt = payload.get("limit", DEFAULT_MAX_FEATURE_COUNT)

    # ensure DEFAULT_PAGE_SIZE independently if custom limit > 500
    if "limit" not in payload or payload["limit"] > DEFAULT_PAGE_SIZE:
        payload["limit"] = DEFAULT_PAGE_SIZE

    page = 1
    features = []
    next_href = None
    with session:
        while True:
            con = _page_search(session, payload, next_href)
            features.extend(con["features"])
            limit_reached = len(features) >= max_cnt
            links = con.get("links", [])

            try:
                next_href = next(filter(lambda c: c["rel"] == "next", links))["href"]
                has_next_page = True
            except StopIteration:
                has_next_page = False
            if not has_next_page or limit_reached:
                break

            page += 1
            logger.info(
                f"\tpage {page} ({(page-1) * payload['limit']} - {page * payload['limit']})"
            )
            payload["page"] = page

    len_feat = len(features)

    if not len_feat:
        logger.info(f"found no STAC items matching your query")
    else:
        multiple_suffix = "s" if len_feat > 1 else ""
        logger.info(f"found {len(features)} STAC item{multiple_suffix}")
    return features


@retry(
    retry_on_exception=retry_if_http_status_error,
    wait_exponential_multiplier=1000,
    stop_max_delay=16000,
)
def _page_search(
    session: CapellaConsoleSession, payload: Dict[str, Any], next_href: str
) -> Dict[str, Any]:
    if next_href is None:
        resp = session.post("/catalog/search", json=payload)
    else:
        resp = session.post(next_href, json=payload)
    body = resp.json()
    return body



from typing import Dict, Any


class CapellaConsoleClientError(Exception):
    response = None
    data: Dict[str, Any] = {}
    message = "An unknown error occurred"

    def __init__(self, message=None, code=None, data={}, response=None):
        self.response = response
        if message:
            self.message = message
        if code:
            self.code = code
        if data:
            self.data = data

    def __str__(self):
        return self.message


class AuthenticationError(CapellaConsoleClientError):
    pass


class NoValidStacIdsError(CapellaConsoleClientError):
    pass


class TaskNotCompleteError(CapellaConsoleClientError):
    pass


class InsufficientFundsError(CapellaConsoleClientError):
    pass


class OrderRejectedError(CapellaConsoleClientError):
    pass


class OrderExpiredError(CapellaConsoleClientError):
    pass


class ConnectError(CapellaConsoleClientError):
    pass


DEFAULT_ERROR_CODE = "GENERAL_API_ERROR"
INVALID_TOKEN_ERROR_CODE = "INVALID_TOKEN"
ORDER_EXPIRED_ERROR_CODE = "ORDER_EXPIRED"

ERROR_CODES = {
    INVALID_TOKEN_ERROR_CODE: AuthenticationError,
    ORDER_EXPIRED_ERROR_CODE: OrderExpiredError,
}

ERROR_CODES_BY_MESSAGE_SNIP = {"order expired": ORDER_EXPIRED_ERROR_CODE}


def handle_error_response(resp):
    error = resp.json()
    if isinstance(error, dict):
        if "error" in error:
            error = error["error"]

        message = error.get("message")
        code = error.get("code", DEFAULT_ERROR_CODE)
        data = error.get("data", {})
    else:
        message = error
        code = DEFAULT_ERROR_CODE
        data = {}

    # try to assign some more meaningful exception class by message
    if code == DEFAULT_ERROR_CODE and message:
        try:
            code = next(
                v
                for k, v in ERROR_CODES_BY_MESSAGE_SNIP.items()
                if k in message.lower()
            )
        except StopIteration:
            pass

    exc = ERROR_CODES.get(code, CapellaConsoleClientError)(
        message=message, code=code, data=data, response=resp
    )
    raise exc



import uuid

from typing import no_type_check, Optional, List, Dict, Any


@no_type_check
def _validate_uuid(uuid_str: str) -> None:
    try:
        uuid.UUID(uuid_str)
    except ValueError as e:
        raise ValueError(f"{uuid_str} is not a valid uuid: {e}")


def _validate_stac_id_or_stac_items(
    stac_ids: Optional[List[str]] = None, items: Optional[List[Dict[str, Any]]] = None
) -> List[str]:
    if not stac_ids and not items:
        raise ValueError("Please provide stac_ids or items")

    if not stac_ids:
        stac_ids = [f["id"] for f in items]  # type: ignore

    return stac_ids



import logging

from .exceptions import handle_error_response
from capella_console_client.exceptions import CapellaConsoleClientError

logger = logging.getLogger()


def translate_error_to_exception(response):
    if response.is_error:
        handle_error_response(response)


def log_on_4xx_5xx(response):
    try:
        response.raise_for_status()
    except Exception:
        request = response.request
        logger.error(
            f"response event hook: {request.method} {request.url} - Status {response.status_code}"
        )
        logger.error(f"the following error occured: {response.json()}")


def retry_if_http_status_error(exception):
    """Return upon httpx.HTTPStatusError"""
    return isinstance(exception, CapellaConsoleClientError)
